#!/usr/bin/perl

# Copyright 2003 by Benjamin Crowell. This software is copyleft licensed under
# the same terms as Perl, or, at your option, under the GPL 1.1 license.

use strict;
use locale;
use utf8;
use Getopt::Long; # Comes with the Perl distribution.

#----------------------------------------------------------------
# Defaults for the preferences:
#----------------------------------------------------------------

our %preferences=(
  'language'=>'en',                 # user's language
  'past'=>-1,                       # how many days into the past the report extends
  'future'=>14,                     # ...and how far into the future
  'calendar'=>'~/.when/calendar',   # where to find the calendar file
  'wrap'=>80,                       # 0 means don't wrap; otherwise, wrap display to this many columns
  'editor'=>'emacs -nw',            # editor
  'filter_accents_on_output'=>1,    # since most Unix terminals show accented Unicode chars as garbage
);

our %command_line_options = (
  'language=s'=>\$preferences{'language'},
  'past=i'=>\$preferences{'past'},
  'future=i'=>\$preferences{'future'},
  'calendar=s'=>\$preferences{'calendar'},
  'wrap=i'=>\$preferences{'wrap'},
  'editor=s'=>\$preferences{'editor'},
  'filter_accents_on_output!'=>\$preferences{'filter_accents_on_output'},
);

#----------------------------------------------------------------
# Strings are all collected here for ease of internationalization:
#----------------------------------------------------------------

our $e_acute = "\x{e9}";

our %month_name =
  (
   'en'=>'Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec',   # English
   'fr'=>"Jan F${e_acute}v Mar Avr Mai Juin Juil Aug Sep Oct Nov D${e_acute}c", # French
  );
our %month_name_long =
  (
    'en'=>'January February March April May June July August September October November December', # English
  );
our %wday_name = ('en'=>'Sun Mon Tue Wed Thu Fri Sat');

sub w {
  my $what = shift;
  my @stuff = @_;
  my $lingo = $preferences{'language'};
  my %strings;
  my $found_lingo = 0;
  if (!$found_lingo) { # default to English
    $found_lingo = 1;
    %strings = 
      (
        'error_opening_prefs'=>"The preferences file %s exists, but there was an error opening it for input.",
        'syntex_err_in_prefs'=>"Syntax error in preferences file %s:\n%s",
        'prefs_file_not_found'=>"Preferences file %s not found.\n",
        'illegal_command'=>"Illegal command, %s\n",
        'error_opening_calendar'=>"Couldn't open the file %s for input\n",
        'yesterday'=>'yesterday',
        'today'=>'today',
        'tomorrow'=>'tomorrow',
        'syntax_error_in_calendar'=>"Syntax error in calendar: %s\n",
        'illegal_year'=>"Illegal year: %s\n",
        'illegal_month'=>"Illegal month: %s\n",
        'illegal_day_of_month'=>"Illegal day of the month: %s\n",
      )
    }
  if (exists $strings{$what}) {$what = $strings{$what}}
  return sprintf($what,@stuff);
}


#----------------------------------------------------------------
# Read the preferences file.
#----------------------------------------------------------------

my $prefs_file = glob '~/.when/preferences';
if (-e $prefs_file) {
  open (FILE,"<$prefs_file") or die w('error_opening_prefs',$prefs_file);
  while (my $line = <FILE>) {
    if ($line =~ m/^\s*(\w+)\s*\=\s*([^\s]*)\s*/) {
      $preferences{lc($1)} = $2;
    }
    else {
      if (!($line =~ m/^\s*$/)) {die w('syntex_err_in_prefs',$prefs_file,$line)}
    }
  }
  close FILE;
}
else {
  do_output(w('prefs_file_not_found',$prefs_file));
}
my $file = glob $preferences{'calendar'};


#----------------------------------------------------------------
# Some global variables.
#----------------------------------------------------------------
our $date_delimiter = ' '; # e.g. 2003 Feb 1
our $use_month_names = 1;
our $use_am_pm = 1;

our @month_length =      (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);

#----------------------------------------------------------------
# Figure out what the user wants to do.
#----------------------------------------------------------------

my $cmds = '';

foreach my $arg(@ARGV) {
  if (! ($arg =~ m/^\-/)) {
    $cmds = $cmds . lc($arg);
  }
}

GetOptions(%command_line_options); # from Getops::Long

my @do_what = ();

if ($cmds eq '') {$cmds = 'di'}

my %periods = ('w'=>7,'m'=>31,'y'=>366);
#loop over chars in arg:
while ($cmds =~ m/(.)/g) {
  my $cmd = $1;
  my $recognized = 0;
  if (exists $periods{$cmd}) {
    $recognized = 1;
    push @do_what,'normal';
    $preferences{'future'} = $periods{$cmd};
  }
  else {
    if ($cmd eq 'd') {
      $recognized = 1;
      push @do_what,'date';
    }
    if ($cmd eq 'i') {
      $recognized = 1;
      push @do_what,'normal';
    }
    if ($cmd eq 'c') {
      $recognized = 1;
      push @do_what,'calendar';
    }
    if ($cmd eq 'e') {
      $recognized = 1;
      push @do_what,'edit';
    }
  }
  if (!$recognized) {
    die w('illegal_command',$cmd);
  }
}

#----------------------------------------------------------------
# Do it.
#----------------------------------------------------------------

my $now = When::current_date();

# For testing:
#    $now = (When::parse_blank_delimited("2003 Jan 30"))->[1];


my $first_one = 1;
foreach my $cmd(@do_what) {
  if (!$first_one || ($first_one && $preferences{'header'})) {
    do_output("\n");
  }
  $first_one = 0;
  if ($cmd eq 'date') {
    my $describe_wday = $now->day_of_week_name;
    my $describe_today = $now->string_human();
    do_output("$describe_wday $describe_today\n");
  }
  if ($cmd eq 'normal') {
    normal_behavior($now);
  }
  if ($cmd eq 'calendar') {
    calendar($now);
  }
  if ($cmd eq 'edit') {
    system($preferences{'editor'}." ".$preferences{'calendar'});
  }
}

sub do_output {
  my $x = shift;
  if ($preferences{'filter_accents_on_output'}) {
    $x = UnicodeTools::filter_out_accents($x);
  }
  print $x;
}

#----------------------------------------------------------------
# The program's normal behavior is to print out all your appointments
# for a certain period (by default, the next two weeks).
#----------------------------------------------------------------

sub normal_behavior {

  my $now = shift->clone;


  open (FILE, "<$file") or die w('error_opening_calendar',$file);
  my @lines = <FILE>; # in array context, returns all lines from file
  close FILE;

  my @show;

  foreach my $line(@lines) {
    chomp $line;
    if ($line =~ m/^\s*([^,#]*)\s*,\s*(([^\s].*)?)/) {
      my $what = $3;
      my $parsed_date = When::parse_blank_delimited($1);
      my ($err,$when) = @$parsed_date;
      if ($err ne '') {
        do_output("$err\n  $line\n");
      }
      if (!$err) {
        if ($when->y eq '*') {
          $when->y($now->y);
          if ($when->delta_days($now)<$preferences{'past'}) {$when->y($when->y+1)}
          if ($when->delta_days($now)>$preferences{'future'}) {$when->y($when->y-1)}
        }
        my $delta = $when->delta_days($now);
        if ($delta >= $preferences{'past'} && $delta <=$preferences{'future'}) {
          my $describe = $when->day_of_week_name;
          if ($delta == -1) {$describe = w('yesterday')}
          if ($delta == 0) {$describe = w('today')}
          if ($delta == 1) {$describe = w('tomorrow')}
          my $say = sprintf "%9s %s %s\n",$describe,$when->string_human,$what;
          push @show,[$when,$say];
        }
      }
    }
    else {
      if (!($line =~ m/^\s*$/ || $line =~ m/^#/)) {die w('syntax_error_in_calendar',$line)}
    }
  }

  @show = sort {$a->[0]->compare($b->[0])} @show;

  my $wrap = $preferences{'wrap'};
  if ($wrap==0) {$wrap=9999}
  my $margin = 22; # This is the width of the column containing the date.
  foreach my $thing(@show) {
    my ($when,$say) = @$thing;
    do_output(format_item($say,$wrap,$margin));
  }
}

#----------------------------------------------------------------
# Print a calendar.
#----------------------------------------------------------------
sub calendar {

  my $now = shift->clone;

  my $last_month = $now->clone;
  $last_month->decrement_month_in_place();
  my @cal1 = make_one_month_calendar(WHEN=>$last_month);

  my @cal2 = make_one_month_calendar(WHEN=>$now,MARK_TODAY=>1);

  my $next_month = $now->clone;
  $next_month->increment_month_in_place();
  my @cal3 = make_one_month_calendar(WHEN=>$next_month);

  my @cals = (\@cal1,\@cal2,\@cal3);

  for (my $line_num=0; $line_num<=10; $line_num++) {
    my $have_one = 0;
    foreach my $cal(@cals) {
      $have_one = $have_one || exists $cal->[$line_num];
    }
    if ($have_one) {
      my $full_line = '';
      foreach my $cal(@cals) {
        my $part;
        if (exists $cal->[$line_num]) {
          $part = $cal->[$line_num];
        }
        else {
          $part = (' ' x 21);  #### constant shouldn't be hardcoded
        }
        if ($full_line ne '') {$full_line = "$full_line  "}
        $full_line = $full_line . $part;
      }
      do_output("$full_line\n");
    }
    else {
      last
    }
  }

}

sub make_one_month_calendar {
  my %args = (
    MARK_TODAY=>0,
    @_,
  );
  my $when = $args{WHEN}->clone;
  my $mark_today = $args{MARK_TODAY};
  my @cal = ();

  my $today = $when->clone;

  my $columns = 21; # 3 columns for each day of the week

  my $month_name_long = When::month_name_long($when->m);
  # Pad it on the front and back so it's roughly centered:
  my $pad_character = '-';
  while (length($month_name_long)<$columns) {
    if (length($month_name_long)%2==0) {
      $month_name_long = "$month_name_long$pad_character";
    }
    else {
      $month_name_long = "$pad_character$month_name_long";
    }
  }
  push @cal,$month_name_long;

  my $line = '';
  for (my $wday=1; $wday<=7; $wday++) {
    $line = $line . ' '.When::short_wday_name($wday).' ';
  }
  push @cal,$line;

  $line = '';
  $when->d(1);
  for (my $wday=1; $wday<$when->day_of_week && $wday<=7; $wday++) {
    $line = $line . '   ';
  }
  my $this_month = $when->m;
  while ($when->m == $this_month) {
    if ($when->day_of_week==1) {push @cal,$line; $line = ''}
    my $display;
    if ($when->compare($today)==0 && $mark_today) {
      $display = ' *';
    }
    else {
      $display = sprintf('%2d',$when->d);
    }
    $line = "$line$display ";
    $when->increment_day_in_place();
  }
  if ($line ne '') {push @cal,$line; $line = ''}

  # Make sure all the lines are equal in length:
  for (my $i=0; $i<=$#cal; $i++) {
    my $line = $cal[$i];
    while (length($line)<$columns) {
      $line = "$line ";
    }
    $cal[$i] = $line;
  }

  return @cal;
}


#----------------------------------------------------------------
# Routines for wrapping long lines:
#----------------------------------------------------------------

# Format a line of text for output. Wrap long lines nicely.
sub format_item {
  my $text = shift;
  my $wrap = shift;
  my $margin = shift;
  chomp $text;
  my @stuff = split_a_line($text,$wrap,2);
  my $result = (shift @stuff)."\n";
  my @the_rest = ();
  if (@stuff) {
    @the_rest = split_a_line((shift @stuff),$wrap-$margin,9999);
  }
  foreach my $line(@the_rest) {
    $result = $result . (' ' x $margin) . "$line\n";
  }
  return $result;
}

# Split a long line into shorter pieces, preferably at word breaks.
sub split_a_line {
  my $text = shift;
  my $width = shift;
  my $max_pieces = shift;
  if (length($text)<=$width || $max_pieces==1) {return ($text,)}
  if ($text =~ m/^(.{0,$width})(\s+(.*))?$/) {
    return ($1,split_a_line($3,$width,$max_pieces-1))
  }
  # The only reason we'd get to this point is if they have an extremely long line
  # consisting of one extremely long word with no blanks in it. This means we can't
  # break at a word boundary.
  $text =~ m/^(.{0,$width})(.*)$/;
  return ($1,split_a_line($2,$width,$max_pieces-1));
}


#----------------------------------------------------------------
# A When object stores year, month, day, time (ymdt).
# Month is 1..12
# Time is null, or hour, or hour:min; hour is on 24-hour time.
# There are methods for doing calculations with the Gregorian calendar.
#----------------------------------------------------------------

package When;


sub new {
  my $class = shift;
  my $self = {};
  bless($self,$class);
  $self->{Y} = shift;
  $self->{M} = shift;
  $self->{D} = shift;
  if (@_) {
    $self->{T}=shift
  }
  else {
    $self->{T} = '';
  }
  return $self;
}

sub parse_blank_delimited {
  my $date = shift;
  chomp $date;
  my @a = split / /,$date;
  if ($a[0] ne '*' && $a[0]<1900 || $a[0]>2100) {return [main::w('illegal_year',$a[0])]}
  $a[1] = parse_month_name($a[1]);
  if ($a[1] eq '' || $a[1]<1 || $a[1]>12) {return [main::w('illegal_month',$a[1])]}
  if ($a[2]<1 || $a[2]>31) {return [main::w('illegal_day_of_month',$a[2])]}
  return ['',When->new(@a)];
}

# can be number or name
# if it's a name, ignores case, trailing dot, and extra characters after first 3 or 4
sub parse_month_name {
  my $name = shift;
  if ($name =~ m/\d+/) {return $name}
  $name = UnicodeTools::filter_out_accents($name);
  for (my $m=1; $m<=12; $m++) {
    my $n = UnicodeTools::filter_out_accents(month_name($m));
    my $n2 = month_name($m,'en');
    if ($name =~ m/^$n/i ||  $name =~ m/^$n2/i) {return $m}
  }
  return '';
}

sub clone {
  my $self = shift;
  return When->new($self->array);
}

sub array {
  my $self = shift;
  return ($self->y,$self->m,$self->d,$self->t);
}

sub y {
  my $self = shift;
  if (@_) {$self->{Y} = shift}
  return $self->{Y};
}

sub m {
  my $self = shift;
  if (@_) {$self->{M} = shift}
  return $self->{M};
}

sub d {
  my $self = shift;
  if (@_) {$self->{D} = shift}
  return $self->{D};
}

sub t {
  my $self = shift;
  if (@_) {$self->{T} = shift}
  return $self->{T};
}

sub hour {
  my $self = shift;
  my $t = $self->t;
  $t =~ m/^\d+/;
  return $1;
}

# returns null string if not set
sub min {
  my $self = shift;
  my $t = $self->t;
  if ($t =~ m/\d+\:(\d_)/) {
    return $1;
  }
  else {
    return '';
  }
}

sub min_no_null {
  my $self = shift;
  my $m = $self->min;
  if ($m eq '') {$m=0}
  return $m;
}

sub current_date {
    my @tm = localtime;
    my $y = $tm[5];
    my $m = $tm[4]+1;
    my $d = $tm[3];
    if ($y<1900) {$y=$y+1900} # works in Perl 5 and 6
    return When->new($y,$m,$d);
}


sub string_sortable {
  my $self = shift;
  return sprintf "%04d-%02d-%02d %02d:%02d", $self->y,$self->m,$self->d,$self->hour,$self->min_no_null;
}

# y,m,d = numbers, m=1..12; t is h or h:m, on 24-hour time
sub string_human {
  my $when = shift;
  my ($y,$m,$d,$t) = $when->array;
  if ($use_month_names) {$m=month_name($m)}
  if ($d<10) {$d=" $d"}
  my $result = $y.$date_delimiter.$m.$date_delimiter.$d;
  if ($t ne '') {$result = $result . ' '.time_string_human($t)}
  return $result;
}

sub time_string_human {
  my $self = shift;
  my ($h,$m) = ($self->hour,$self->min);
  my $suffix = '';
  if ($use_am_pm) {
    if ($h>12) {
      $h=$h-12;
      $suffix = 'pm'
    }
    else {
      $suffix = 'am';
    }
  }
  if ($m ne '') {
    return sprintf '%d:%02d%s',$h,$m,$suffix;
  }
  else {
    return sprintf '%d%s',$h,$suffix;
  }
}

sub month_name {
  my $m = shift;
  return month_name_short_or_long($m,'short',@_);
}

sub month_name_long {
  my $m = shift;
  return month_name_short_or_long($m,'long',@_);
}

sub month_name_short_or_long {
  my $m = shift;
  my $short_or_long = shift;
  my $list_of_names;
  if ($short_or_long eq 'short') {
    $list_of_names = \%month_name;
  }
  if ($short_or_long eq 'long') {
    $list_of_names = \%month_name_long;
  }
  if (! ref $list_of_names) {return ''}
  my $language = $preferences{'language'};
  if (@_) {$language = shift}
  if ($m<1 || $m>12) {return ''}
  my $names;
  if (exists $list_of_names->{$language}) {
    $names = $list_of_names->{$language}
  }
  else {
    $names = $list_of_names->{'en'};
  }
  my @names = split / /,$names;
  return $names[$m-1];
}

sub wday_name {
  my $d = shift;
  if ($d<1 || $d>7) {return ''}
  my $names = $wday_name{'en'};
  if (exists $wday_name{$preferences{'language'}}) {
    $names = $wday_name{$preferences{'language'}}
  }
  my @names = split / /,$names;
  return $names[$d-1];
}

sub short_wday_name {
  my $d = shift;
  if ($d<1 || $d>7) {return ''}
  wday_name($d) =~ m/^(.)/; # extract first character
  return $1;
}

sub compare {
  my $a = shift;
  my $b = shift;
  if ($a->y != $b->y) {return $a->y <=> $b->y}
  if ($a->m != $b->m) {return $a->m <=> $b->m}
  return $a->d <=> $b->d;
}

sub increment_day_in_place {
  my $self = shift;
  $self->d($self->d+1);
  if ($self->d <= $self->days_in_month()) {return}
  $self->m($self->m+1);
  $self->d(1);
  if ($self->m <= 12) {return}
  $self->y($self->y+1);
  $self->m(1);
}

sub increment_month_in_place {
  my $self = shift;
  $self->m($self->m+1);
  if ($self->m > 12) {
    $self->y($self->y+1);
    $self->m(1);
  }
  while ($self->d > $self->days_in_month()) {
    $self->d($self->d-1)
  }
}

sub decrement_month_in_place {
  my $self = shift;
  $self->m($self->m-1);
  if ($self->m < 1) {
    $self->y($self->y-1);
    $self->m(12);
  }
  while ($self->d > $self->days_in_month()) {
    $self->d($self->d-1)
  }
}

sub delta_days {
  my $a = shift;
  my $b = shift;
  my $compared = $a->compare($b);
  if ($compared == 0) {return 0}
  if ($compared == -1) {return -($b->delta_days($a))}
  if ($a->d != 1 || $b->d !=1) {
    my $aa = $a->clone;
    my $bb = $b->clone;
    $aa->{D} = 1;
    $bb->{D} = 1;
    return ($aa->delta_days($bb))+($a->d)-($b->d);
  }
  if ($a->m != 1 || $b->m !=1) {
    my $aa = $a->clone;
    my $bb = $b->clone;
    my $correction = 0;
    while ($aa->m > 1) {
      $aa->m($aa->m-1);
      $correction = $correction + $aa->days_in_month;
    }
    while ($bb->m > 1) {
      $bb->m($bb->m-1);
      $correction = $correction - $bb->days_in_month;
    }
    return ($aa->delta_days($bb))+$correction;
  }
  # From Jan 1 of one year to Jan 1 of another; $a is after $b
    my $aa = $a->clone;
    my $result = 0;
    while ($aa->y > $b->y) {
      $aa->y($aa->y-1);
      if ($aa->is_leap_year) {
        $result = $result+366;
      }
      else {
        $result = $result+365;
      }
    }
    return $result;
}

sub days_in_month {
  my $self = shift;
  if ($self->m != 2) {return $month_length[($self->m)-1]}
  if ($self->is_leap_year) {
    return 29;
  }
  else {
    return 28;
  }
}

sub is_leap_year {
  my $self = shift;
  my $y = $self->y;
  if ($y%4!=0) {return 0}
  if ($y%100!=0) {return 1}
  if ($y%1000!=0) {return 0}
  return 1;
}

sub day_of_week {
  my $self =shift;
  return ($self->delta_days(When->new(2003,2,2))%7)+1;
}

sub day_of_week_name {
  my $self =shift;
  return wday_name($self->day_of_week);
}

#----------------------------------------------------------------
# Defaults for the preferences:
#----------------------------------------------------------------

package UnicodeTools;

sub filter_out_accents {
  my $x = shift;
  my %filter = ($e_acute=>'e');
  while ($x =~ m/([^a-zA-Z0-9])/g) {
    my $a = $1;
    if (exists $filter{$a}) {
      $b = $filter{$a};
      $x =~ s/$a/$b/g;
    }
  }
  return $x;
}
